//
// gpsr_receiver.cc : GPSR Receiver Main File
// author           : Xi Wang
// note             : Converted from ping_receiver
//
// Copyright (C) 2000-2003 by the University of Southern California
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include "gpsr_receiver.hh"

#ifdef NS_DIFFUSION
static class GPSRReceiverAppClass : public TclClass {
public:
    GPSRReceiverAppClass() : TclClass("Application/DiffApp/GPSRReceiver") {}
    TclObject* create(int , const char*const* ) {
	    return(new GPSRReceiverApp());
    }
} class_ping_receiver;

int GPSRReceiverApp::command(int argc, const char*const* argv) {
  if (argc == 2) {
    if (strcmp(argv[1], "subscribe") == 0) {
      run();
      return TCL_OK;
    }
  }
  else if (argc == 3) {
    if (strcmp(argv[1], "trace") == 0) {
      trace_en = atoi(argv[2]);
      return TCL_OK;
    }
    else if (strcmp(argv[1], "id") == 0) {
      my_id_ = atoi(argv[2]);
      return TCL_OK;
    }
  }
  return DiffApp::command(argc, argv);
}

#endif // NS_DIFFUSION

void GPSRReceiverReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  app_->recv(data, my_handle);
}

void GPSRReceiverApp::recv(NRAttrVec *data, NR::handle my_handle)
{
  NRSimpleAttribute<int> *counterAttr = NULL;
  NRSimpleAttribute<void *> *timeAttr = NULL;
  GpsrEventTime *probe_event;
  long delay_seconds;
  long delay_useconds;
  float total_delay;
  struct timeval tmv;

  GetTime(&tmv);

  counterAttr = GpsrCounterAttr.find(data);
  timeAttr = GpsrTimeAttr.find(data);

  if (!counterAttr || !timeAttr){
    DiffPrint(DEBUG_ALWAYS, "Received a BAD packet !\n");
    PrintAttrs(data);
    return;
  }

  // Calculate latency
  probe_event = (GpsrEventTime *) timeAttr->getVal();
  delay_seconds = tmv.tv_sec;
  delay_useconds = tmv.tv_usec;

  if ((delay_seconds < probe_event->seconds_) ||
      ((delay_seconds == probe_event->seconds_) &&
       (delay_useconds < probe_event->useconds_))){

    // Time's not synchronized
    delay_seconds = -1;
    delay_useconds = 0;
    DiffPrint(DEBUG_ALWAYS, "Error calculating delay !\n");
  }
  else{
    delay_seconds = delay_seconds - probe_event->seconds_;
    if (delay_useconds < probe_event->useconds_){
      delay_seconds--;
      delay_useconds = delay_useconds + 1000000;
    }
    delay_useconds = delay_useconds - probe_event->useconds_;
  }
  total_delay = (float) (1.0 * delay_seconds) + ((float) delay_useconds / 1000000.0);

  // Check if this is the first message received
  if (first_msg_recv_ < 0){
    first_msg_recv_ = counterAttr->getVal();
  }

  // Print output message
  if (last_seq_recv_ >= 0){
    if (counterAttr->getVal() < last_seq_recv_){
      // Multiple sources detected, disabling statistics
      last_seq_recv_ = -1;
      DiffPrint(DEBUG_ALWAYS, "Received data %d, total latency = %f!\n",
		counterAttr->getVal(), total_delay);
    }
    else{
      last_seq_recv_ = counterAttr->getVal();
      num_msg_recv_++;
      DiffPrint(DEBUG_ALWAYS, "Received data: %d, total latency = %f, %% messages received: %f !\n",
		last_seq_recv_, total_delay,
		(float) ((num_msg_recv_ * 100.00) /
			 ((last_seq_recv_ - first_msg_recv_) + 1)));
    }
  }
  else{
    DiffPrint(DEBUG_ALWAYS, "Received data %d, total latency = %f !\n",
	      counterAttr->getVal(), total_delay);
  }

#ifdef NS_DIFFUSION
  if (trace_en) {
    printf(
	   "#DR "TIME_FORMAT" %d %d %f\n",
	   Scheduler::instance().clock(),
	   my_id_,
	   last_seq_recv_, 
	   total_delay
	   );
  }
#endif
}

handle GPSRReceiverApp::setupSubscription()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS,
				   NRAttribute::INTEREST_CLASS));
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS,
				   NRAttribute::NODE_LOCAL_SCOPE));
  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS,
				       NRAttribute::GPSR_ALGORITHM));
  attrs.push_back(LongitudeAttr.make(NRAttribute::IS, my_x)); 
  attrs.push_back(LatitudeAttr.make(NRAttribute::IS, my_y)); 
  attrs.push_back(GpsrTargetAttr.make(NRAttribute::IS, "F117A"));

  handle h = dr_->subscribe(&attrs, mr_);

  ClearAttrs(&attrs);

  return h;
}

void GPSRReceiverApp::run()
{
#ifdef NS_DIFFUSION
  // get location from ns
  double z;
  MobileNode *node = ((DiffusionRouting *)dr_)->getNode();
  node->getLoc(&my_x, &my_y, &z);
#endif

  subHandle_ = setupSubscription();

#ifndef NS_DIFFUSION
  // Do nothing
  while (1){
    sleep(1000);
  }
#endif // !NS_DIFFUSION
}

#ifdef NS_DIFFUSION
GPSRReceiverApp::GPSRReceiverApp()
#else
GPSRReceiverApp::GPSRReceiverApp(int argc, char **argv)
#endif // NS_DIFFUSION
{
  last_seq_recv_ = 0;
  num_msg_recv_ = 0;
  first_msg_recv_ = -1;

  my_x = 0;
  my_y = 0;

  my_id_ = 0;
  trace_en = 0;

  mr_ = new GPSRReceiverReceive(this);

#ifndef NS_DIFFUSION
  parseCommandLine(argc, argv);
  dr_ = NR::createNR(diffusion_port_);
#endif // !NS_DIFFUSION
}

#ifndef NS_DIFFUSION
int main(int argc, char **argv)
{
  GPSRReceiverApp *app;

  app = new GPSRReceiverApp(argc, argv);
  app->run();

  return 0;
}
#endif // !NS_DIFFUSION
