//
// gpsr_sender.cc : GPSR Sender main file
// author         : Xi Wang
// note           : Converted from ping_sender
//
// Copyright (C) 2000-2003 by the University of Southern California
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include "gpsr_sender.hh"
#include <unistd.h>

#ifdef NS_DIFFUSION
static class GPSRSenderAppClass : public TclClass {
public:
  GPSRSenderAppClass() : TclClass("Application/DiffApp/GPSRSender") {}
  TclObject* create(int , const char*const*) {
    return(new GPSRSenderApp());
  }
} class_gpsr_sender;

int GPSRSenderApp::command(int argc, const char*const* argv) {
  if (argc == 2) {
    if (strcmp(argv[1], "publish") == 0) {
      run();
      return TCL_OK;
    }
  }
  else if (argc == 3) {
    if (strcmp(argv[1], "debug") == 0) {
      global_debug_level = atoi(argv[2]);
      return TCL_OK;
    }
    else if (strcmp(argv[1], "publish") == 0) {
      num_sends = atoi(argv[2]);
      run();
      return TCL_OK;
    }
    else if (strcmp(argv[1], "trace") == 0) {
      trace_en = atoi(argv[2]);
      return TCL_OK;
    }
    else if (strcmp(argv[1], "id") == 0) {
      my_id_ = atoi(argv[2]);
      return TCL_OK;
    }
  }  else if (argc == 5) {
    // TCL API $app algo destination x y
    if (strcmp(argv[1], "gpsr") == 0) {
      algo = GPSR;
    }
    else if (strcmp(argv[1], "gear") == 0) {
      algo = GEAR;
    }
    else if (strcmp(argv[1], "push") == 0) {
      algo = PUSH;
    }
    else {
      return DiffApp::command(argc, argv);
    }

    if (strcmp(argv[2], "destination") == 0) {
      dest_x = atof(argv[3]);
      dest_y = atof(argv[4]);
      mode = GPSRSenderApp::EXACT;
      return TCL_OK;
    }
    else if (strcmp(argv[2], "near") == 0) {
      dest_x = atof(argv[3]);
      dest_y = atof(argv[4]);
      mode = GPSRSenderApp::NEAR;
      return TCL_OK;
    }
    else {
      return DiffApp::command(argc, argv);
    }
  }
  return DiffApp::command(argc, argv);
}

void GPSRSendDataTimer::expire(Event *e) {
  a_->send();
}

void GPSRSenderApp::send()
{
  int retval;

  if (num_sends != 0 && last_seq_sent_ >= num_sends) {
    return;
  }

  // Send data probe
  DiffPrint(DEBUG_ALWAYS, "Sending Data %d\n", last_seq_sent_);
  retval = dr_->send(pubHandle_, &data_attr_);

  if (trace_en) {
    printf(
	   "#SS "TIME_FORMAT" %d %d\n",
	   Scheduler::instance().clock(),
	   my_id_,
	   last_seq_sent_
	   );
  }

  // Update counter
  last_seq_sent_++;
  counterAttr_->setVal(last_seq_sent_);

  // re-schedule the timer 
  sdt_.resched(SEND_DATA_INTERVAL);
}
#endif // NS_DIFFUSION

void GPSRSenderReceive::recv(NRAttrVec *data, NR::handle my_handle)
{
  app_->recv(data, my_handle);
}

void GPSRSenderApp::recv(NRAttrVec *data, NR::handle my_handle)
{
  NRSimpleAttribute<int> *nrclass = NULL;

  nrclass = NRClassAttr.find(data);

  if (!nrclass){
    DiffPrint(DEBUG_ALWAYS, "Received a BAD packet !\n");
    return;
  }

  switch (nrclass->getVal()){

  case NRAttribute::INTEREST_CLASS:

    DiffPrint(DEBUG_ALWAYS, "Received an Interest message !\n");
    num_subscriptions_++;
    break;

  case NRAttribute::DISINTEREST_CLASS:

    DiffPrint(DEBUG_ALWAYS, "Received a Disinterest message !\n");
    num_subscriptions_--;
    break;

  default:

    DiffPrint(DEBUG_ALWAYS, "Received an unknown message (%d)!\n", nrclass->getVal());
    break;

  }
}


handle GPSRSenderApp::setupPublication()
{
  NRAttrVec attrs;

  attrs.push_back(NRClassAttr.make(NRAttribute::IS, NRAttribute::DATA_CLASS));
  attrs.push_back(NRScopeAttr.make(NRAttribute::IS, NRAttribute::GLOBAL_SCOPE));
  if (algo == GPSR) {
    attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS, NRAttribute::GPSR_ALGORITHM));
    if (mode == GPSRSenderApp::EXACT) {
      attrs.push_back(LongitudeAttr.make(NRAttribute::EQ, dest_x));
      attrs.push_back(LatitudeAttr.make(NRAttribute::EQ, dest_y));
    }
    else if (mode == GPSRSenderApp::NEAR) { 
      attrs.push_back(LongitudeAttr.make(NRAttribute::EQ_CLOSEST, dest_x));
      attrs.push_back(LatitudeAttr.make(NRAttribute::EQ_CLOSEST, dest_y));
    }
  }
  else if (algo == GEAR) {
    attrs.push_back(LongitudeAttr.make(NRAttribute::EQ, dest_x));
    attrs.push_back(LatitudeAttr.make(NRAttribute::EQ, dest_y));    
  }
  else if (algo == PUSH) {
    attrs.push_back(LongitudeAttr.make(NRAttribute::EQ, dest_x));
    attrs.push_back(LatitudeAttr.make(NRAttribute::EQ, dest_y));    
  }
  else {
    return NULL;
  }

  attrs.push_back(GpsrTargetAttr.make(NRAttribute::IS, "F117A"));

  handle h = dr_->publish(&attrs);

  ClearAttrs(&attrs);

  return h;
}

void GPSRSenderApp::run()
{
  struct timeval tmv;
#ifndef NS_DIFFUSION
  int retval;
#endif // !NS_DIFFUSION

#ifdef INTERACTIVE
  char input;
  fd_set FDS;
#endif // INTERATIVE

  // Setup publication and subscription
  pubHandle_ = setupPublication();

  // Create time attribute
  GetTime(&tmv);
  lastEventTime_ = new GpsrEventTime;
  lastEventTime_->seconds_ = tmv.tv_sec;
  lastEventTime_->useconds_ = tmv.tv_usec;
  timeAttr_ = GpsrTimeAttr.make(NRAttribute::IS, (void *) &lastEventTime_,
				sizeof(GpsrEventTime));
  data_attr_.push_back(timeAttr_);

  // Change pointer to point to attribute's payload
  delete lastEventTime_;
  lastEventTime_ = (GpsrEventTime *) timeAttr_->getVal();

  // Create counter attribute
  counterAttr_ = GpsrCounterAttr.make(NRAttribute::IS, last_seq_sent_);
  data_attr_.push_back(counterAttr_);

#ifndef NS_DIFFUSION
  // Main thread will send data probes
  while(1){
#ifdef INTERACTIVE
    FD_SET(0, &FDS);
    fprintf(stdout, "Press <Enter> to send a data probe...");
    fflush(NULL);
    select(1, &FDS, NULL, NULL, NULL);
    input = getc(stdin);
#else
    sleep(SEND_DATA_INTERVAL);
#endif // INTERACTIVE

    // Update time in the packet
    GetTime(&tmv);
    lastEventTime_->seconds_ = tmv.tv_sec;
    lastEventTime_->useconds_ = tmv.tv_usec;

    // Send data probe
    DiffPrint(DEBUG_ALWAYS, "Sending Data %d\n", last_seq_sent_);
    retval = dr_->send(pubHandle_, &data_attr_);

    // Update counter
    last_seq_sent_++;
    counterAttr_->setVal(last_seq_sent_);
  }
#else
  send();
#endif // !NS_DIFFUSION
}

#ifdef NS_DIFFUSION
GPSRSenderApp::GPSRSenderApp() : sdt_(this)
#else
GPSRSenderApp::GPSRSenderApp(int argc, char **argv)
#endif // NS_DIFFUSION
{
  last_seq_sent_ = 0;
  num_subscriptions_ = 0;

  dest_x = 0;
  dest_y = 0;
  num_sends = 0;
  algo = GPSR;

  my_id_ = 0;
  trace_en = 0;

  mr_ = new GPSRSenderReceive(this);

#ifndef NS_DIFFUSION
  parseCommandLine(argc, argv);
  dr_ = NR::createNR(diffusion_port_);
#endif // NS_DIFFUSION
}

#ifndef NS_DIFFUSION
void GPSRSenderApp::parseCommandLine(int argc, char **argv)
{
  u_int16_t diff_port = DEFAULT_DIFFUSION_PORT;
  int debug_level;
  int opt;

  config_file_ = NULL;
  opterr = 0;

  while (1){
    opt = getopt(argc, argv, "h:d:p:x:y:");
    switch (opt){

    case 'x':

      dest_x = atof(optarg);
      break;

    case 'y':

      dest_y = atof(optarg);
      break;

    case 'p':

      diff_port = (u_int16_t) atoi(optarg);
      if ((diff_port < 1024) || (diff_port >= 65535)){
	DiffPrint(DEBUG_ALWAYS, "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 'h':

      usage(argv[0]);

      break;

    case 'd':

      debug_level = atoi(optarg);

      if (debug_level < 1 || debug_level > 10){
	DiffPrint(DEBUG_ALWAYS, "Error: Debug level outside range or missing !\n");
	usage(argv[0]);
      }

      global_debug_level = debug_level;

      break;

    case '?':

      DiffPrint(DEBUG_ALWAYS,
		"Error: %c isn't a valid option or its parameter is missing !\n",
		optopt);
      usage(argv[0]);

      break;

    case ':':

      DiffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage(argv[0]);

      break;

    }

    if (opt == -1)
      break;
  }

  diffusion_port_ = diff_port;
}

int main(int argc, char **argv)
{
  GPSRSenderApp *app;

  app = new GPSRSenderApp(argc, argv);
  app->run();

  return 0;
}
#endif // NS_DIFFUSION

