//
// gpsr.cc        : GPSR Filter main file
// author         : Xi Wang
// note           : Contains code pieces from - 
//                  Source Routing Filter 
//                  GEAR Filter
//
// Copyright (C) 2000-2003 by the Unversity of Southern California
// $Id: gpsr.cc,v 1.1 2003/09/18 22:54:15 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include "gpsr.hh"
#include "gpsr_attr.hh"

#ifdef NS_DIFFUSION
class DiffAppAgent;
#endif // NS_DIFFUSION

#ifdef NS_DIFFUSION
static class GPSRFilterClass : public TclClass {
public:
  GPSRFilterClass() : TclClass("Application/DiffApp/GPSRFilter") {}
  TclObject* create(int argc, const char*const* argv) {
    if (argc == 5) {
      return (new GPSRFilter(atoi(argv[4])));
    }
    fprintf(stderr, "Wrong number of args for creating GPSR Filter\n");
    return NULL;
    //return(new GPSRFilter());
  }
} class_gpsr_filter;


int GPSRFilter::command(int argc, const char*const* argv) {
  Tcl& tcl =  Tcl::instance();
  if (argc == 2) {
    if (strcmp(argv[1], "start") == 0) {
      run();
      return (TCL_OK);
    }
  } 
  else if (argc == 3) {
    if (strcmp(argv[1], "log") == 0) {
      if (strcmp(argv[2], "on") == 0) {
	log_en = 1;
	return (TCL_OK);
      }
      else if (strcmp(argv[2], "off") == 0) {
	log_en = 0;
	return (TCL_OK);
      }
    }
  }

  return (DiffApp::command(argc, argv));
}
#endif // NS_DIFFUSION

// under construction, logging function
void GPSRFilter::Log(int type, GPSRHeader &hdr) 
{
  if (!log_en) {
    return;
  }
  
  char types[10];
  if (type > 0) {
    sprintf(types, "recv");
  }
  else if (type < 0) {
    sprintf(types, "drop");
  }
  else {
    return;
  }

  printf("%s %d %d\n", types, my_id_, hdr.hopcount);
}

double GPSRFilter::GeoDist(double x1, double y1, double x2, double y2)
{
  return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
}

GPSRNeighborEntry * GPSRFilter::findNeighbor(int32_t neighbor_id)
{
  GPSRNeighborList::iterator neighbor_itr;
  GPSRNeighborEntry *neighbor_entry;
 
  for (neighbor_itr = neighbors.begin();
       neighbor_itr != neighbors.end(); ++neighbor_itr){
    neighbor_entry = *neighbor_itr;
    if (neighbor_entry->nid == neighbor_id) {
      return neighbor_entry;
    }
 }
 return NULL;
}

void GPSRFilter::updateNeighbor(int32_t neighbor_id, double location_x,
				double location_y)
{
  GPSRNeighborEntry *neighbor_entry;

  // Look for this neighbor in our neighbor's list
  neighbor_entry = findNeighbor(neighbor_id);

  if (!neighbor_entry){
    // Insert a new neighbor into our list
    DiffPrintWithTime(DEBUG_IMPORTANT,
		      "GPSR-%d: Inserting new neighbor %d. \n",
		      my_id_, neighbor_id);

    neighbor_entry = new GPSRNeighborEntry;
    neighbor_entry->nid = neighbor_id;
    neighbor_entry->location_x = location_x;
    neighbor_entry->location_y = location_y;
    neighbor_entry->ttl = GPSR_NEIGHBOR_TTL;

    // Insert new element with key neighbor_id into the hash map
    neighbors.push_back(neighbor_entry);
  }
  else{
    // Update an existing neighbor entry
    DiffPrintWithTime(DEBUG_LOTS_DETAILS,
		      "GPSR - Updating neighbor %d.\n", neighbor_id);

    neighbor_entry->location_x = location_x;
    neighbor_entry->location_y = location_y;
    neighbor_entry->ttl = GPSR_NEIGHBOR_TTL;
  }
}

void GPSRFilterReceive::recv(Message *msg, handle h)
{
  app_->recv(msg, h);
}

void GPSRFilter::recv(Message *msg, handle h)
{
  Message *return_msg = NULL;

  if (h != filter_handle){
    DiffPrint(DEBUG_ALWAYS,
	      "Error: Received a message for handle %ld when subscribing to handle %ld !\n", h, filter_handle);
    return;
  }

  return_msg = ProcessMessage(msg);

  if (return_msg){
    ((DiffusionRouting *)dr_)->sendMessage(msg, h);

    delete msg;
  }
}

void GPSRFilter::DataToSink(Message *msg)
{
  Message * sink_message = CopyMessage(msg);
  GPSRSinkList::iterator sink_itr;
  GPSRSinkEntry *sink_entry;

  for (sink_itr = sink_list.begin(); sink_itr != sink_list.end(); sink_itr++){
    sink_entry = *sink_itr;

    if (MatchAttrs(sink_entry->attrs, msg->msg_attr_vec_)){
      // Forward DATA to local sinks
      sink_message->next_hop_ = LOCALHOST_ADDR;
      sink_message->next_port_ = sink_entry->port;
      ((DiffusionRouting *)dr_)->sendMessage(sink_message, FILTER_MIN_PRIORITY);
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Message delivered to sink %d\n",
		my_id_, sink_entry->port);
    }
  }

  delete sink_message;
}

int GPSRFilter::SinkExpire()
{
  GPSRSinkList::iterator sink_itr;
  GPSRSinkEntry *sink_entry;
 
  for (sink_itr = sink_list.begin(); sink_itr != sink_list.end(); sink_itr++){
    sink_entry = *sink_itr;
    sink_entry->ttl -= GPSR_SINKTIMER_INTERVAL;

    if (sink_entry->ttl <= 0){
      DiffPrintWithTime(DEBUG_IMPORTANT,
			"GPSR-%d: Sink %d expired, removing..\n",
			my_id_, sink_entry->port);
      sink_itr = sink_list.erase(sink_itr);
    }
  }
  return 0;
}

int GPSRFilter::FindLoc(NRAttrVec *attrlist, double & x, double & y)
{
  NRSimpleAttribute<float> * locx = LongitudeAttr.find(attrlist);
  NRSimpleAttribute<float> * locy = LatitudeAttr.find(attrlist);

  if (locx && locy){
    if (locx->getOp() == NRAttribute::EQ &&
	locy->getOp() == NRAttribute::EQ){
      x = locx->getVal();
      y = locy->getVal();
      return 1;
    }
    else
      if (locx->getOp() == NRAttribute::IS &&
	  locy->getOp() == NRAttribute::IS) {
	x = locx->getVal();
	y = locy->getVal();
	return 1;
      }
      else
	if (locx->getOp() == NRAttribute::EQ_CLOSEST &&
	    locy->getOp() == NRAttribute::EQ_CLOSEST) {
	  x = locx->getVal();
	  y = locy->getVal();
	  return 2;
	}
	else{
	  return 0;
	}
  }
  else{
    return 0;
  }
}

Message * GPSRFilter::ProcessMessage(Message *msg)
{
  SGPSRBeaconAttr * beaconattr;
  int32_t neighbor_id;

  neighbor_id = msg->last_hop_;
  if ((neighbor_id == BROADCAST_ADDR)){// || (neighbor_id == LOCALHOST_ADDR)){ We want to receive locally generated test messages
    DiffPrintWithTime(DEBUG_ALWAYS, "GPSR-%d: Neighbor ID Invalid.\n", my_id_);
    return NULL;
  }

  NRSimpleAttribute<int> * nrclass;
  NRSimpleAttribute<int> * nrscope; 
  int found;
  GPSRSinkList::iterator sink_itr;
  GPSRSinkEntry *sink_entry;
  
  switch (msg->msg_type_){

  case INTEREST:
 
    DiffPrint(DEBUG_ALWAYS, "GPSR-%d: Interest received.\n", my_id_);

    nrclass = NRClassAttr.find(msg->msg_attr_vec_);
    nrscope = NRScopeAttr.find(msg->msg_attr_vec_);

    if (!nrclass || !nrscope){
      DiffPrint(DEBUG_ALWAYS,
		"GPSR-%d Warning: Can't find CLASS/SCOPE attributes in the message.\n", my_id_);
      return NULL;
    }
    
    if (msg->last_hop_ != LOCALHOST_ADDR){
      // Interest received from the network.
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d Warning: Received interest from network, dropping.\n",
		my_id_);
      return NULL;
    }
    
    found = 0;
    for (sink_itr = sink_list.begin();sink_itr != sink_list.end();sink_itr++) {
      sink_entry = *sink_itr;
      if (sink_entry->port == msg->source_port_) {
	found = 1;
	break;
      }
    }

    if (!found){
      // Create a new routing entry for this data type
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Adding sink entry %d.\n", my_id_, msg->source_port_);
      sink_entry = new GPSRSinkEntry;
      sink_entry->attrs = CopyAttrs(msg->msg_attr_vec_);
      sink_entry->port = msg->source_port_;
      sink_entry->ttl = GPSR_SINK_TTL;
      sink_list.push_back(sink_entry);
    }
    else {
      sink_entry->ttl = GPSR_SINK_TTL;
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Updating sink entry %d.\n", my_id_, msg->source_port_);
    }
    break;

  case PUSH_EXPLORATORY_DATA:
    DiffPrint(DEBUG_IMPORTANT,
	      "GPSR-%d: Received push exploratory data, converting to normal data.\n", my_id_);
    msg->msg_type_ = DATA;
  case DATA:
    DiffPrint(DEBUG_LOTS_DETAILS, "GPSR-%d: Data message received.\n", my_id_);

    beaconattr = GPSRBeaconAttr.find(msg->msg_attr_vec_);
    if (!beaconattr) { // normal data

      double vlx, vly;
      int type = FindLoc(msg->msg_attr_vec_, vlx, vly);

      if (type != 0){

	SGPSRHeaderAttr *pgh = GPSRHeaderAttr.find(msg->msg_attr_vec_);

	if (!pgh){

	  // Add default header for new messages
	  GPSRHeader tmpgh;
	  tmpgh.mode = GPSRH_DATA_GREEDY;
	  tmpgh.hopcount = 0;
	  tmpgh.deliverymode = (type==1)? GPSRH_EXACTDELIVERY : GPSRH_CLOSEDELIVERY;
	  tmpgh.cnode = 0;
	  tmpgh.cdist = -1;
	  char *buf = new char[tmpgh.Size()];
	  tmpgh.Pack(buf);
	  msg->msg_attr_vec_->push_back(GPSRHeaderAttr.make(NRAttribute::IS,
							    buf, tmpgh.Size()));
	  delete buf;
	  pgh = GPSRHeaderAttr.find(msg->msg_attr_vec_);
	  DiffPrintWithTime(DEBUG_IMPORTANT, "GPSR-%d: ****** GPSR Header added for new packet, start forwarding ****\n", my_id_);
	}

	GPSRHeader gpsrheader;
	gpsrheader.Unpack(pgh->getVal());

	if ((vlx == location_x && vly == location_y) || GeoDist(vlx, vly, location_x, location_y) <= GPSR_TARGETERR) {

	  // This is the destination. We use both exact match and proxy
	  // match to avoid numerical precision problems

	  DiffPrintWithTime(DEBUG_IMPORTANT, "GPSR-%d: ********************* Packet reached destination ****\n", my_id_);
	  DataToSink(msg);
	  Log(1, gpsrheader);
	}

	else
	  if (gpsrheader.deliverymode == GPSRH_TOCLOSE &&
	      gpsrheader.cnode == my_id_){
	    DiffPrintWithTime(DEBUG_IMPORTANT,
			      "GPSR-%d: ********************* Packet close delivered to this node  ****\n", my_id_);
	    DataToSink(msg);
	    Log(2, gpsrheader);
	  }
	  else{
	    // forward
	    gpsrheader.hopcount++;
	    if (gpsrheader.hopcount >= GPSR_MAXHOP){
	      DiffPrint(DEBUG_IMPORTANT,
			"***************** GPSR-%d: Hop limit reached, dropping. ****\n", my_id_);
	      Log(-2, gpsrheader);
	    }
	    else{

	      if (gpsrheader.deliverymode == GPSRH_CLOSEDELIVERY){
	      
		// Maintain closest known node
		double dist = distance(location_x, location_y, vlx, vly);
		if (dist < gpsrheader.cdist || gpsrheader.cdist == -1){
		  gpsrheader.cdist = dist;
		  gpsrheader.cnode = my_id_;
		}
	      }

	      int reason;
	      GPSRNeighborEntry * ety = Routing(&gpsrheader, vlx, vly, (int)(msg->last_hop_), reason);
	      if (ety){
	      
		// forwarding possible
		msg->next_hop_ = ety->nid;
		gpsrheader.Pack(pgh->getVal());
		((DiffusionRouting *)dr_)->sendMessage(msg, filter_handle,
						       FILTER_MIN_PRIORITY);
		DiffPrint(DEBUG_IMPORTANT,
			  "**** GPSR-%d: Message forwarded to %d. ****\n",
			  my_id_, ety->nid);
	      }
	      else{

		// Process close delivery
		if (gpsrheader.deliverymode == GPSRH_CLOSEDELIVERY &&
		    reason == GPSR_ROUTING_NOEDGE){

		  // should deliver to myself
		  DiffPrintWithTime(DEBUG_IMPORTANT,
				    "GPSR-%d: ********************* No outside edge, deliver to myself ****\n", my_id_);
		  DataToSink(msg);
		  Log(3, gpsrheader);
		}
		else
		  if (gpsrheader.deliverymode == GPSRH_CLOSEDELIVERY &&
		      reason == GPSR_ROUTING_CIRCLE){
		    DiffPrint(DEBUG_IMPORTANT,
			      "GPSR-%d: Re-enter the circle for close delivery\n",
			      my_id_);
		    gpsrheader.deliverymode = GPSRH_TOCLOSE;
		    // restart perimeter mode
		    double brg = bearing(location_x, location_y, vlx, vly);
		    // find neighbor with greatest bearing <= brg
		    GPSRNeighborEntry *ne = neighbors.ent_next_ccw(brg, location_x, location_y, use_planar_, NULL);
		    if (!ne){

		      // no edge
		      DiffPrint(DEBUG_IMPORTANT,
				"GPSR: No edge from this node, dropping\n");
		      Log(-4, gpsrheader);
		    }
		    else{
		      msg->next_hop_ = ne->nid;
		      gpsrheader.Pack(pgh->getVal());
		      ((DiffusionRouting *)dr_)->sendMessage(msg, filter_handle,
							     FILTER_MIN_PRIORITY);
		      DiffPrint(DEBUG_IMPORTANT,
				"**** GPSR-%d: Message forwarded to %d. ****\n",
				my_id_, ne->nid);
		    }
		  }
		  else
		    if (gpsrheader.deliverymode == GPSRH_TOCLOSE){
		      DiffPrint(DEBUG_IMPORTANT, "***************** GPSR-%d: Unable to find closest node previously indetified, dropping. ****\n", my_id_);
		      Log(-2, gpsrheader);
		    }
		    else{

		      // We just drop for exact delivery
		      DiffPrint(DEBUG_IMPORTANT, "***************** GPSR-%d: Dropping unroutable packet. ****\n", my_id_);
		      Log(-1, gpsrheader);
		    }
	      }
	    }
	  }
      }
      else{
	DiffPrint(DEBUG_ALWAYS,
		  "GPSR-%d: GPSR message missing location attributes !\n",
		  my_id_);
      }
    }

    else{

      // received beacon
      DiffPrint(DEBUG_LOTS_DETAILS,
		"GPSR-%d: Beacon message received.\n", my_id_);

      NRSimpleAttribute<float> * locx = LongitudeAttr.find(msg->msg_attr_vec_);
      NRSimpleAttribute<float> * locy = LatitudeAttr.find(msg->msg_attr_vec_);

      if (locx && locy){
	double vlx = locx->getVal();
	double vly = locy->getVal();
	updateNeighbor(neighbor_id, vlx, vly);
	neighbors.planarize(PLANARIZE_GABRIEL, location_x, location_y);
      }
      else{
	DiffPrint(DEBUG_ALWAYS,
		  "GPSR-%d: No location attributes found.\n", my_id_);
      }
    }

    break;

  case EXPLORATORY_DATA:

    DiffPrint(DEBUG_IMPORTANT,
	      "GPSR-%d Warning: Received exploratory data.\n", my_id_);
    break;

    /* case PUSH_EXPLORATORY_DATA:
    DiffPrint(DEBUG_IMPORTANT, "GPSR-%d Warning: Received push exploratory data.\n", my_id_);
    break;*/

  case  POSITIVE_REINFORCEMENT:

    DiffPrint(DEBUG_IMPORTANT,
	      "GPSR-%d Warning: Received positive reinforcement.\n", my_id_);
    break;

  case NEGATIVE_REINFORCEMENT:

    DiffPrint(DEBUG_IMPORTANT,
	      "GPSR-%d Warning: Received negative reinforcement.\n", my_id_);
    break;

  default:

    DiffPrint(DEBUG_ALWAYS,
	      "GPSR-%d Warning: Unknown message received.\n", my_id_);
  }

  return NULL;
}

int GPSRFilter::BeaconExpire()
{
  NRAttrVec attrs;
  Message *msg;

  DiffPrintWithTime(DEBUG_LOTS_DETAILS,
		    "GPSR - Beacon timer expired, sending beacon packet.\n");
  
  //send beacon message
  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS,
				       NRAttribute::GPSR_ALGORITHM));
  attrs.push_back(GPSRBeaconAttr.make(NRAttribute::IS, 0));
  attrs.push_back(LongitudeAttr.make(NRAttribute::IS, location_x));
  attrs.push_back(LatitudeAttr.make(NRAttribute::IS, location_y));

  msg = new Message(DIFFUSION_VERSION, DATA, 0, 0, attrs.size(),
		    pkt_count_, rdm_id_, BROADCAST_ADDR, LOCALHOST_ADDR);

  msg->msg_attr_vec_ = CopyAttrs(&attrs);

  pkt_count_++;

  ((DiffusionRouting *)dr_)->sendMessage(msg, FILTER_MIN_PRIORITY);

  ClearAttrs(&attrs);
  delete msg;

  int rto, to;
  if (beacon_timer_count < GPSR_BEACON_P1REP) {
    // Faster initial rate
    to = GPSR_BEACON_INTERVAL1;
    beacon_timer_count++;
  }
  else {
    // Slower rate
    to = GPSR_BEACON_INTERVAL2;
  }

  rto = (int) ((float)to*(1-GPSR_BEACON_RANDRATIO) + (float)rand()/(float)RAND_MAX*(float)to*GPSR_BEACON_RANDRATIO*2.0);

  return rto;
}

int GPSRFilter::NeighborExpire()
{
  GPSRNeighborList::iterator neighbor_itr;
  GPSRNeighborEntry *neighbor_entry;
 
  for (neighbor_itr = neighbors.begin();
       neighbor_itr != neighbors.end(); neighbor_itr++){

    neighbor_entry = *neighbor_itr;
    neighbor_entry->ttl -= GPSR_NTIMER_INTERVAL;

    if (neighbor_entry->ttl <= 0){

      DiffPrintWithTime(DEBUG_IMPORTANT,
			"GPSR-%d: Neighbor %d expired, removing..\n",
			my_id_, neighbor_entry->nid);
      neighbor_itr = neighbors.erase(neighbor_itr);
    }
  }
  return 0;
}

int GPSRFilter::TestSendExpire()
{
  NRAttrVec attrs;
  Message *msg;

  if (beacon_timer_count < 4){
    return 0;
  }

  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::IS,
				       NRAttribute::GPSR_ALGORITHM));
  attrs.push_back(LongitudeAttr.make(NRAttribute::IS, testdest_x));
  attrs.push_back(LatitudeAttr.make(NRAttribute::IS, testdest_y));

  msg = new Message(DIFFUSION_VERSION, DATA, 0, 0, attrs.size(),
		    pkt_count_, rdm_id_, LOCALHOST_ADDR, LOCALHOST_ADDR);

  msg->msg_attr_vec_ = CopyAttrs(&attrs);

  pkt_count_++;

  ((DiffusionRouting *)dr_)->sendMessage(msg, filter_handle,
					 GPSR_FILTER_PRIORITY + 1);

  // send to myself
  DiffPrintWithTime(DEBUG_ALWAYS,
		    "****************** GPSR-%d: Test message sent. ****\n",
		    my_id_);

  ClearAttrs(&attrs);
  delete msg; 

  return 0;
}

handle GPSRFilter::setupFilter()
{
  NRAttrVec attrs;
  handle h;

  // Match all packets with a GPSR Attribute
  attrs.push_back(NRAlgorithmAttr.make(NRAttribute::EQ,
				       NRAttribute::GPSR_ALGORITHM));

  h = ((DiffusionRouting *)dr_)->addFilter(&attrs, GPSR_FILTER_PRIORITY,
					   filter_callback_);

  ClearAttrs(&attrs);
  return h;
}

void GPSRFilter::run()
{
  int rto, to;

  // Add periodic beacon timer
  to = GPSR_BEACON_INTERVAL1;
  rto = (int) ((float)to*(1.0-GPSR_BEACON_RANDRATIO) + (float)rand()/(float)RAND_MAX*(float)to*GPSR_BEACON_RANDRATIO*2.0);

  beacon_timer = new GPSRBeaconTimer(this);
  ((DiffusionRouting *)dr_)->addTimer(rto, beacon_timer);
  beacon_timer_count = 0;
  
  nexpire_timer = new GPSRNExpireTimer(this);
  ((DiffusionRouting *)dr_)->addTimer(GPSR_NTIMER_INTERVAL, nexpire_timer);

  sinkexpire_timer = new GPSRSinkExpireTimer(this);
  ((DiffusionRouting *)dr_)->addTimer(GPSR_SINKTIMER_INTERVAL, sinkexpire_timer);

  // Test timer 
  if (TestInterval != 0){
    test_timer = new GPSRTestTimer(this);
    ((DiffusionRouting *)dr_)->addTimer(TestInterval, test_timer);
  }
   
#ifdef NS_DIFFUSION
  // get location from ns
  double z;
  MobileNode *node = ((DiffusionRouting *)dr_)->getNode();
  node->getLoc(&location_x, &location_y, &z);

  filter_handle = setupFilter();
  DiffPrint(DEBUG_ALWAYS,
	    "GPSRFilter filter received handle %ld\n",  filter_handle);
  DiffPrint(DEBUG_ALWAYS, "GPSRFilter filter initialized !\n");
#else
  while (1){
    sleep(1000);
  }
#endif // NS_DIFFUSION
}

#ifdef NS_DIFFUSION
GPSRFilter::GPSRFilter(int id)
{
  DiffAppAgent *agent;
  my_id_ = id;
#else
GPSRFilter::GPSRFilter(int argc, char **argv)
{
#endif // NS_DIFFUSION
  struct timeval tv;

  location_x = 0;
  location_y = 0;
  TestInterval = 0;
  testdest_x = 0;
  testdest_y = 0;
  use_planar_= 1;
  log_en = 0;

  // Create Diffusion Routing class
#ifndef NS_DIFFUSION
  parseCommandLine(argc, argv);
  dr_ = NR::createNR(diffusion_port_);
#endif // !NS_DIFFUSION

  filter_callback_ = new GPSRFilterReceive(this);

  GetTime(&tv);
  SetSeed(&tv);
  pkt_count_ = GetRand();
  rdm_id_ = GetRand();

#ifndef NS_DIFFUSION
  // Set up the filter
  filter_handle = setupFilter();
  DiffPrint(DEBUG_ALWAYS, "GPSR filter received handle %ld\n", filter_handle);
  DiffPrint(DEBUG_ALWAYS, "GPSR filter initialized !\n");
#endif // !NS_DIFFUSION
}

#ifndef NS_DIFFUSION
// Location command line parameters are added
void GPSRFilter::parseCommandLine(int argc, char **argv)
{
  u_int16_t diff_port = DEFAULT_DIFFUSION_PORT;
  int debug_level;
  char *nidenv;
  int opt;

  config_file_ = NULL;
  opterr = 0;

  while (1){
    //opt = getopt(argc, argv, "f:hd:p:x:y:s:z:w:");
    opt = getopt(argc, argv, "h:d:p:x:y:s:z:w:");
    switch (opt){

    case 'x':
      location_x = atof(optarg);
      break;

    case 'y':
      location_y = atof(optarg);
      break;

    case 's':
      TestInterval = atoi(optarg);
      break;

    case 'z':
      testdest_x = atof(optarg);
      break;

    case 'w':
      testdest_y = atof(optarg);
      break;

    case 'p':

      diff_port = (u_int16_t) atoi(optarg);
      if ((diff_port < 1024) || (diff_port >= 65535)){
	DiffPrint(DEBUG_ALWAYS,
		  "Error: Diffusion port must be between 1024 and 65535 !\n");
	exit(-1);
      }

      break;

    case 'h':

      usage(argv[0]);

      break;

    case 'd':

      debug_level = atoi(optarg);

      if (debug_level < 1 || debug_level > 10){
	DiffPrint(DEBUG_ALWAYS,
		  "Error: Debug level outside range or missing !\n");
	usage(argv[0]);
      }

      global_debug_level = debug_level;

      break;

    case '?':

      DiffPrint(DEBUG_ALWAYS,
		"Error: %c isn't a valid option or its parameter is missing !\n",
		optopt);
      usage(argv[0]);

      break;

    case ':':

      DiffPrint(DEBUG_ALWAYS, "Parameter missing !\n");
      usage(argv[0]);

      break;

    }

    if (opt == -1)
      break;
  }

  diffusion_port_ = diff_port;

  // Get node id from environment variable
  nidenv = getenv("node_addr");

  if (nidenv){
    my_id_ = atoi(nidenv);
  }
  else{
    DiffPrint(DEBUG_ALWAYS, "GPSR: Node ID envirment variable not found.\n");
    my_id_ = 0;
  }
}
#endif

#ifndef USE_SINGLE_ADDRESS_SPACE

int main(int argc, char **argv)
{
  GPSRFilter *app;

  // Initialize and run the Source Route Filter
  app = new GPSRFilter(argc, argv);
  app->run();

  return 0;
}
#endif // !USE_SINGLE_ADDRESS_SPACE
