//
// gpsr_core.cc   : GPSR core algorithms
// author         : Xi Wang
// note           : Coverted from Brad Karp's ns GPSR implementation
//
// Copyright (C) 2000-2003 by the Unversity of Southern California
// $Id: gpsr_core.cc,v 1.1 2003/09/18 22:54:15 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>

#include "gpsr.hh"
#include "gpsr_attr.hh"

inline double distance(double x1, double y1,  double x2, double y2)
{
  return sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2));
}

inline double bearing(double x1, double y1, double x2, double y2)
{
  double brg;

  brg = atan2(y2 - y1, x2 - x1);
  if (brg < 0)
    brg += 2*M_PI;
  return brg;
}

int GPSRNeighborEntry::closer_pt(int myid, double myx, double myy,
			 double ptx, double pty,
			 int  ptipa, int  ptipb,
			 double dstx, double dsty,
			 double *closerx, double *closery)
{
  if ((min(nid, myid) == min(ptipa, ptipb)) &&
      (max(nid, myid) == max(ptipa, ptipb))) {
    // this edge is the same edge where (ptx, pty) lies; nope
    return 0;
  }
  if (!live) {
    // this edge is not part of the planarized graph
    return 0;
  }
  if (cross_segment(ptx, pty, dstx, dsty, myx, myy,
		    location_x, location_y, closerx, closery)) {
    if (distance(*closerx, *closery, dstx, dsty) < distance(ptx, pty, dstx, dsty)) {
      // edge has point closer than (ptx, pty)
      return 1;
    }
  }
  return 0;
}

GPSRNeighborEntry * GPSRNeighborList::ent_findshortest(double myx, double myy,
						       double tx, double ty)
{
  double shortest, t;
  GPSRNeighborList::iterator neighbor_itr;
  GPSRNeighborEntry * ne = NULL, * cne;

  shortest = distance(myx, myy, tx, ty);
  for (neighbor_itr = begin(); neighbor_itr != end(); neighbor_itr++){
    cne = *neighbor_itr;
    if ((t = distance(cne->location_x, cne->location_y, tx, ty)) < shortest) {
      shortest = t;
      ne = cne;
    }
  }
  return ne;
}


GPSRNeighborEntry * GPSRNeighborList::ent_next_ccw(double basebrg, double x,
						   double y, int p,
						   GPSRNeighborEntry *inne)
{
  GPSRNeighborEntry *minne = NULL, *ne;
  double brg, minbrg = 3*M_PI;

  GPSRNeighborList::iterator neighbor_itr;

  for (neighbor_itr = begin(); neighbor_itr != end(); neighbor_itr++){
    ne = *neighbor_itr;
    if (inne && (ne == inne))
      continue;
    if (p && !ne->live)
      continue;
    brg = bearing(x, y, ne->location_x, ne->location_y) - basebrg;
    if (brg < 0)
      brg += 2*M_PI;
    if (brg < 0)
      brg += 2*M_PI;
    if (brg < minbrg) {
      minbrg = brg;
      minne = ne;
    }
  }
  return minne;
}

GPSRNeighborEntry * GPSRNeighborList::ent_next_ccw(double myx, double myy,
						   GPSRNeighborEntry *inne,
						   int p)
{
  double brg;
  GPSRNeighborEntry *ne;

  brg = bearing(myx, myy, inne->location_x, inne->location_y);
  ne = ent_next_ccw(brg, myx, myy, p, inne);
  if (!ne)
    return inne;
  else
    return ne;
}


GPSRNeighborEntry *GPSRNeighborList::ent_findbyid(int nid)
{
  GPSRNeighborList::iterator neighbor_itr;
  GPSRNeighborEntry * neighbor_entry, * r = NULL;

  for (neighbor_itr = begin(); neighbor_itr != end(); neighbor_itr++){
    neighbor_entry = *neighbor_itr;
    if (neighbor_entry->nid == nid) {
      r = neighbor_entry;
      break;
    }
  }
  return r;
}


void GPSRNeighborList::planarize_one(GPSRNeighborEntry *nep, int algo,
				     double x, double y)
{ 
  double uvdist, canddist, midx, midy;

  GPSRNeighborList::iterator neighbor_itr;
  GPSRNeighborEntry *nei;

  uvdist = distance(x, y, nep->location_x, nep->location_y);
  switch(algo) {
  case PLANARIZE_RNG:
    break;
  case PLANARIZE_GABRIEL:
    // find midpt of segment me (u) <-> this (v)
    midx = (x + nep->location_x) / 2.0;
    midy = (y + nep->location_y) / 2.0;
    uvdist /= 2.0;
    break;
  default:
    DiffPrint(DEBUG_ALWAYS,
	      "GPSR - Programming error - Unknown planarization algorithm \n");
    break;
  }
  for (neighbor_itr = begin(); neighbor_itr != end(); neighbor_itr++){
    nei = *neighbor_itr;
    if (nep == nei) {
      // w and v identical node--w not a witness
      continue;
    }
    switch(algo) {

    case PLANARIZE_RNG:
      // find max dist. from me (u) to ne (w) vs. this (v) to ne (w)
      canddist = max(distance(x, y, nei->location_x, nei->location_y),
		     distance(nep->location_x, nep->location_y, nei->location_x, nei->location_y));
      // is max < dist from me (u) to this (v)?
      if (canddist < uvdist) {
	nep->live = 0;
	return;
      }
      break;

    case PLANARIZE_GABRIEL:
      // is ne (w) inside circle of radius uvdist?
      if (distance(midx, midy, nei->location_x, nei->location_y) < uvdist) {
	nep->live = 0;
	return;
      }
      break;

    default:
      break;
    }
  }
  nep->live = 1;
}

void GPSRNeighborList::planarize(int algo, double x, double y)
{
  GPSRNeighborList::iterator neighbor_itr;

  for (neighbor_itr = begin(); neighbor_itr != end(); neighbor_itr++){
    planarize_one(*neighbor_itr, algo, x, y);
  }
}


inline int cross_segment(double x1, double y1, double x2, double y2,
			 double x3, double y3, double x4, double y4,
			 double *xi, double *yi)
{
  double dy[2], dx[2], m[2], b[2];
  double xint, yint;

  dy[0] = y2 - y1;
  dx[0] = x2 - x1;
  dy[1] = y4 - y3;
  dx[1] = x4 - x3;
  m[0] = dy[0] / dx[0];
  m[1] = dy[1] / dx[1];
  b[0] = y1 - m[0] * x1;
  b[1] = y3 - m[1] * x3;
  if (m[0] != m[1]) {
    // slopes not equal, compute intercept
    xint = (b[0] - b[1]) / (m[1] - m[0]);
    yint = m[1] * xint + b[1];
    // is intercept in both line segments?
    if ((xint <= max(x1, x2)) && (xint >= min(x1, x2)) &&
	(yint <= max(y1, y2)) && (yint >= min(y1, y2)) &&
	(xint <= max(x3, x4)) && (xint >= min(x3, x4)) &&
	(yint <= max(y3, y4)) && (yint >= min(y3, y4))) {
      if (xi && yi) {
	*xi = xint;
	*yi = yint;
      }
      return 1;
    }
  }
  return 0;
}


GPSRNeighborEntry * GPSRFilter::Routing(GPSRHeader *gpsrh, double tlocx,
					double tlocy, int lasthop, int &reason) 
{
  GPSRNeighborEntry *ne;

  reason = 0;

  switch(gpsrh->mode) {

  case GPSRH_DATA_GREEDY:
    ne = neighbors.ent_findshortest(location_x, location_y, tlocx, tlocy);
    if (ne) {
      // Greedy forward
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Continue forwarding in greedy mode.\n", my_id_);
      return ne;
    }
    else {
      // find bearing to dst
      double brg = bearing(location_x, location_y, tlocx, tlocy);
      // find neighbor with greatest bearing <= brg
      ne = neighbors.ent_next_ccw(brg, location_x, location_y, use_planar_, NULL);
      if (!ne) {
	// no face toward the destination
	DiffPrint(DEBUG_IMPORTANT, "GPSR: No edge from this node.\n");
	reason = GPSR_ROUTING_NOEDGE;
	return NULL;
      }

      // put packet in peri data mode, forward
      gpsrh->mode = GPSRH_DATA_PERI;
      // mark point of entry into peri data mode
      gpsrh->enterperi_x = location_x;
      gpsrh->enterperi_y = location_y;
      // mark ips of edge endpoints
      gpsrh->perihead[0] = my_id_; // myself
      gpsrh->perihead[1] = ne->nid; // next edge on peri

      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Switching to perimeter mode.\n", my_id_);
      return ne;
    }
    break;

  case GPSRH_DATA_PERI:
    double closerx, closery;

    // non-source-routed perimeter forwarding rule
    /* to resume greedy forwarding, this *node* must be closer than the point where the packet entered peri mode. */
    if (distance(location_x, location_y, tlocx, tlocy) < distance(gpsrh->enterperi_x, gpsrh->enterperi_y, tlocx, tlocy)) {
      gpsrh->mode = GPSRH_DATA_GREEDY;
      // recursive
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Switching back to greedy mode.\n", my_id_);
      return Routing(gpsrh, tlocx, tlocy, lasthop, reason);
    }
    // forward along current face, or change faces where appropriate
    /* don't choose *any* edge--only consider edges on the face we're forwarding on at the moment. */
    ne = neighbors.ent_findbyid(lasthop);
    if (ne){
      GPSRNeighborEntry *nexte = neighbors.ent_next_ccw(location_x, location_y, ne, use_planar_);
      if (!nexte){
	// turn back
	nexte = ne;
	DiffPrint(DEBUG_IMPORTANT, "GPSR-%d: Perimeter mode u-turn.\n", my_id_);
      }

      /* drop if we've looped on this perimeter: are about to revisit
	 the first edge we took on it */
      if ((gpsrh->perihead[0] == my_id_) && (gpsrh->perihead[1] == nexte->nid)){
	DiffPrint(DEBUG_IMPORTANT,
		  "GPSR-%d: Perimeter mode completed a circle - no route.\n",
		  my_id_);
	reason = GPSR_ROUTING_CIRCLE;
	return NULL;
      }
      // does the candidate next edge have a closer pt?
      if (nexte->closer_pt(my_id_, location_x, location_y, gpsrh->enterperi_x, gpsrh->enterperi_y, gpsrh->perihead[0], gpsrh->perihead[1], tlocx, tlocy, &closerx, &closery)) {
	// switch face
	/* choose a new next hop on the peri pierced by the line  to the destination. */
	/* several neighboring edges may be cut by line to destination; choose that cut at the point closest to destination */
	while(nexte->closer_pt(my_id_, location_x, location_y, gpsrh->enterperi_x, gpsrh->enterperi_y, gpsrh->perihead[0], gpsrh->perihead[1], tlocx, tlocy, &closerx, &closery)) {
	  // record closest point on edge to ne
	  gpsrh->enterperi_x = closerx;
	  gpsrh->enterperi_y = closery;
	  nexte = neighbors.ent_next_ccw(location_x, location_y, nexte, use_planar_);
	} 
	// record edge endpt ips
	gpsrh->perihead[0] = my_id_; // self
	gpsrh->perihead[1] = nexte->nid; // next hop
	DiffPrint(DEBUG_IMPORTANT,
		  "GPSR-%d: Switching face in perimeter mode.\n", my_id_);
	return nexte;
      }
      else {
	// continue on current face
	DiffPrint(DEBUG_IMPORTANT,
		  "GPSR-%d: Continue forwarding in perimeter mode.\n", my_id_);
	return nexte;
      }
    }
    else {
      DiffPrint(DEBUG_IMPORTANT,
		"GPSR-%d: Last Hop not in neighbor list - asymmetric link?\n",
		my_id_);
      reason = GPSR_ROUTING_HOPNOTFOUND;
      return NULL;
    }

  default:

    DiffPrint(DEBUG_ALWAYS, "GPSR: Programming error - unknown forward mode.\n");
    reason = GPSR_ROUTING_UNKNOWNMODE;
    return NULL;
    break;
  }
  return NULL; // shouldn't reach here
}


